@@ -1,13 +1,13 @@
 /*
  *  SQLRemoveTranslator.c
  *
- *  $Id$
+ *  $Id: SQLRemoveTranslator.c,v 1.7 2006/01/20 15:58:35 source Exp $
  *
  *  These functions intentionally left blank
  *
  *  The iODBC driver manager.
  *
- *  Copyright (C) 1996-2023 OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2006 by OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -77,7 +77,7 @@
 
 #include <iodbc.h>
 #include <odbcinst.h>
-#include "unicode.h"
+#include <unicode.h>
 
 #include "inifile.h"
 #include "misc.h"
